@@ -1,8 +1,8 @@
-(** A failure occured. *)
-exception Failure
-(** A failure occured (with given explanation). *)
-exception Failure_msg of string
+(** An error occured (with given explanation). *)
+exception Error of string
 exception Timeout
+exception Stopped
+exception Failed
 (** Trying to read data from a stream which has ended. *)
 exception End_of_stream
 
@@ -95,6 +95,9 @@ module Element : sig
   (** Current position of an element. *)
   val position : t -> Format.t -> Int64.t
 
+  (** Duration of an element. *)
+  val duration : t -> Format.t -> Int64.t
+
   (** Seek to a given position relative to the start of the stream. *)
   val seek_simple : t -> Format.t -> Event.seek_flag list -> Int64.t -> unit
 end
@@ -106,35 +109,100 @@ module Element_factory : sig
   val make : string -> string -> t
 end
 
-(** Messages. *)
-module Message : sig
-  type message_type =
-  | Error
-  | Tag
-  | State_changed
-  | Stream_status
-  | Duration_changed
-  | Async_done
-  | Stream_start
-
+(** Main loop. *)
+module Loop : sig
   type t
-
-  val message_type : t -> message_type
-
-  val source_name : t -> string
-
-  val parse_tag : t -> (string * string list) list
+  
+  val create : unit -> t
+  val run : t -> unit
+  val quit : t -> unit
 end
 
 (** Buses. *)
 module Bus : sig
   type t
 
+  type message_type = [
+    | `Unknown
+    | `End_of_stream
+    | `Error
+    | `Warning
+    | `Info
+    | `Tag
+    | `Buffering
+    | `State_changed
+    | `State_dirty
+    | `Step_done
+    | `Clock_provide
+    | `Clock_lost
+    | `New_clock
+    | `Structure_change
+    | `Stream_status
+    | `Application
+    | `Element
+    | `Segment_start
+    | `Segment_done
+    | `Duration_changed
+    | `Latency
+    | `Async_start
+    | `Async_done
+    | `Request_state
+    | `Step_start
+    | `Qos
+    | `Progress
+    | `Toc
+    | `Reset_time
+    | `Stream_start
+    | `Need_context
+    | `Have_context
+    | `Any
+  ]
+
+  type message_payload = [
+    | `Unknown
+    | `End_of_stream
+    | `Error of string
+    | `Warning of string
+    | `Info of string
+    | `Tag of (string*string list) list
+    | `Buffering of int
+    | `State_changed of (Element.state*Element.state*Element.state)
+    | `State_dirty
+    | `Step_done
+    | `Clock_provide
+    | `Clock_lost
+    | `New_clock
+    | `Structure_change
+    | `Stream_status
+    | `Application
+    | `Element
+    | `Segment_start
+    | `Segment_done
+    | `Duration_changed
+    | `Latency
+    | `Async_start
+    | `Async_done
+    | `Request_state
+    | `Step_start
+    | `Qos
+    | `Progress
+    | `Toc
+    | `Reset_time
+    | `Stream_start
+    | `Need_context
+    | `Have_context
+  ]
+
+  type message = {
+    source:  string;
+    payload: message_payload
+  }
+
   val of_element : Element.t -> t
 
-  val pop_filtered : t -> Message.message_type list -> Message.t option
+  val pop_filtered : t -> message_type list -> message option
 
-  val timed_pop_filtered : t -> Message.message_type list -> Message.t
+  val timed_pop_filtered : t -> ?timeout:Int64.t -> message_type list -> message
 end
 
 (** Bins. *)
@@ -194,8 +262,11 @@ module App_src : sig
   (** Push a buffer. *)
   val push_buffer : t -> Buffer.t -> unit
 
-  (** Push a buffer in string format. *)
-  val push_buffer_string : t -> string -> unit
+  (** Push a buffer in bytes format. *)
+  val push_buffer_bytes : t -> ?presentation_time:Int64.t -> ?duration:Int64.t -> bytes -> int -> int -> unit
+
+  (** Push a buffer in data format. *)
+  val push_buffer_data : t -> ?presentation_time:Int64.t -> ?duration:Int64.t -> data -> int -> int -> unit
 
   (** Register a callback that will be called when data need to be fed into the
       source (the argument is the number of bytes needed by the source). *)
