@@ -7,16 +7,46 @@
 #include <caml/bigarray.h>
 #include <caml/threads.h>
 
+#ifndef Bytes_val
+#define Bytes_val String_val
+#endif
+
 #include <string.h>
 #include <assert.h>
 #include <stdio.h>
 
+#include <glib.h>
 #include <gst/gst.h>
 #include <gst/gstclock.h>
 #include <gst/gsttypefind.h>
 #include <gst/app/gstappsrc.h>
 #include <gst/app/gstappsink.h>
 
+#include <pthread.h>
+
+static pthread_key_t ocaml_c_thread_key;
+static pthread_once_t ocaml_c_thread_key_once = PTHREAD_ONCE_INIT;
+
+static void ocaml_gstreamer_on_thread_exit(void *key) {
+  caml_c_thread_unregister();
+}
+
+static void ocaml_gstreamer_make_key() {
+  pthread_key_create(&ocaml_c_thread_key, ocaml_gstreamer_on_thread_exit);
+}
+
+static void ocaml_gstreamer_register_thread() {
+  static int initialized = 1;
+  void *ptr;
+  
+  pthread_once(&ocaml_c_thread_key_once, ocaml_gstreamer_make_key);
+
+  if ((ptr = pthread_getspecific(ocaml_c_thread_key)) == NULL) {
+    pthread_setspecific(ocaml_c_thread_key,(void*)&initialized);
+    caml_c_thread_register();
+  }
+}
+
 CAMLprim value ocaml_gstreamer_init(value _argv)
 {
   CAMLparam1(_argv);
@@ -138,13 +168,10 @@ static struct custom_operations element_ops =
     custom_deserialize_default
   };
 
-static value value_of_element(GstElement *e)
-{
-  if (!e) caml_raise_not_found();
-  value ans = caml_alloc_custom(&element_ops, sizeof(GstElement*), 0, 1);
-  Element_val(ans) = e;
-  return ans;
-}
+#define value_of_element(e,ans) \
+  if (!e) caml_raise_not_found(); \
+  ans = caml_alloc_custom(&element_ops, sizeof(GstElement*), 0, 1); \
+  Element_val(ans) = e; \
 
 CAMLprim value ocaml_gstreamer_element_set_property_string(value e, value l, value v)
 {
@@ -191,7 +218,7 @@ static value value_of_state_change_return(GstStateChangeReturn ret)
   switch(ret)
     {
     case GST_STATE_CHANGE_FAILURE:
-      caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
+      caml_raise_constant(*caml_named_value("gstreamer_exn_failed"));
 
     case GST_STATE_CHANGE_SUCCESS:
       return Val_int(0);
@@ -258,7 +285,7 @@ CAMLprim value ocaml_gstreamer_element_link(value _src, value _dst)
   ret = gst_element_link(src, dst);
   caml_acquire_runtime_system();
 
-  if (!ret) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
+  if (!ret) caml_raise_constant(*caml_named_value("gstreamer_exn_failed"));
   CAMLreturn(Val_unit);
 }
 
@@ -274,10 +301,26 @@ CAMLprim value ocaml_gstreamer_element_position(value _e, value _fmt)
   ret = gst_element_query_position(e, fmt, &pos);
   caml_acquire_runtime_system();
 
-  if (!ret) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
+  if (!ret) caml_raise_constant(*caml_named_value("gstreamer_exn_failed"));
   CAMLreturn(caml_copy_int64(pos));
 }
 
+CAMLprim value ocaml_gstreamer_element_duration(value _e, value _fmt)
+{
+  CAMLparam2(_e, _fmt);
+  GstElement *e = Element_val(_e);
+  GstFormat fmt = format_val(_fmt);
+  gint64 dur;
+  gboolean ret;
+
+  caml_release_runtime_system();
+  ret = gst_element_query_duration(e, fmt, &dur);
+  caml_acquire_runtime_system();
+
+  if (!ret) caml_raise_constant(*caml_named_value("gstreamer_exn_failed"));
+  CAMLreturn(caml_copy_int64(dur));
+}
+
 CAMLprim value ocaml_gstreamer_element_seek_simple(value _e, value _fmt, value _flags, value _pos)
 {
   CAMLparam4(_e, _fmt, _flags, _pos);
@@ -295,7 +338,7 @@ CAMLprim value ocaml_gstreamer_element_seek_simple(value _e, value _fmt, value _
   ret = gst_element_seek_simple(e, fmt, flags, pos);
   caml_acquire_runtime_system();
 
-  if (!ret) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
+  if (!ret) caml_raise_constant(*caml_named_value("gstreamer_exn_failed"));
   CAMLreturn(Val_unit);
 }
 
@@ -308,15 +351,49 @@ CAMLprim value ocaml_gstreamer_element_factory_make(value factname, value name)
   GstElement *e;
 
   e = gst_element_factory_make(String_val(factname), String_val(name));
-  ans = value_of_element(e);
+  value_of_element(e, ans);
 
   CAMLreturn(ans);
 }
 
 /**** Message *****/
 
-#define message_types_len 7
-static const GstMessageType message_types[message_types_len] = { GST_MESSAGE_ERROR, GST_MESSAGE_TAG, GST_MESSAGE_STATE_CHANGED, GST_MESSAGE_STREAM_STATUS, GST_MESSAGE_DURATION_CHANGED, GST_MESSAGE_ASYNC_DONE, GST_MESSAGE_STREAM_START };
+#define message_types_len 33
+static const GstMessageType message_types[message_types_len] = {
+  GST_MESSAGE_UNKNOWN,
+  GST_MESSAGE_EOS,
+  GST_MESSAGE_ERROR,
+  GST_MESSAGE_WARNING,
+  GST_MESSAGE_INFO,
+  GST_MESSAGE_TAG,
+  GST_MESSAGE_BUFFERING,
+  GST_MESSAGE_STATE_CHANGED,
+  GST_MESSAGE_STATE_DIRTY,
+  GST_MESSAGE_STEP_DONE,
+  GST_MESSAGE_CLOCK_PROVIDE,
+  GST_MESSAGE_CLOCK_LOST,
+  GST_MESSAGE_NEW_CLOCK,
+  GST_MESSAGE_STRUCTURE_CHANGE,
+  GST_MESSAGE_STREAM_STATUS,
+  GST_MESSAGE_APPLICATION,
+  GST_MESSAGE_ELEMENT,
+  GST_MESSAGE_SEGMENT_START,
+  GST_MESSAGE_SEGMENT_DONE,
+  GST_MESSAGE_DURATION_CHANGED,
+  GST_MESSAGE_LATENCY,
+  GST_MESSAGE_ASYNC_START,
+  GST_MESSAGE_ASYNC_DONE,
+  GST_MESSAGE_REQUEST_STATE,
+  GST_MESSAGE_STEP_START,
+  GST_MESSAGE_QOS,
+  GST_MESSAGE_PROGRESS,
+  GST_MESSAGE_TOC,
+  GST_MESSAGE_RESET_TIME,
+  GST_MESSAGE_STREAM_START,
+  GST_MESSAGE_NEED_CONTEXT,
+  GST_MESSAGE_HAVE_CONTEXT,
+  GST_MESSAGE_ANY
+};
 
 static GstMessageType message_type_of_int(int n)
 {
@@ -353,13 +430,9 @@ static struct custom_operations message_ops =
     custom_deserialize_default
   };
 
-static value value_of_message(GstMessage *msg)
-{
-  if (!msg) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
-  value ans = caml_alloc_custom(&message_ops, sizeof(GstMessage*), 0, 1);
+#define value_of_message(msg, ans) \
+  ans = caml_alloc_custom(&message_ops, sizeof(GstMessage*), 0, 1); \
   Message_val(ans) = msg;
-  return ans;
-}
 
 CAMLprim value ocaml_gstreamer_message_type(value _msg)
 {
@@ -375,6 +448,44 @@ CAMLprim value ocaml_gstreamer_message_source_name(value _msg)
   CAMLreturn(caml_copy_string(GST_MESSAGE_SRC_NAME(msg)));
 }
 
+CAMLprim value ocaml_gstreamer_message_parse_error(value _err)
+{
+  CAMLparam1(_err);
+  CAMLlocal1(ans);
+  GError *err = NULL;
+
+  gst_message_parse_error(Message_val(_err), &err, NULL);
+
+  ans = caml_copy_string(err->message);
+  g_error_free(err);
+  
+  CAMLreturn(ans);
+}
+
+CAMLprim value ocaml_gstreamer_message_parse_buffering(value _msg)
+{
+  CAMLparam1(_msg);
+  gint percent;
+  gst_message_parse_buffering(Message_val(_msg), &percent);
+  CAMLreturn(Val_int(percent));
+}
+
+CAMLprim value ocaml_gstreamer_message_parse_state_changed(value _msg)
+{
+  CAMLparam1(_msg);
+  CAMLlocal1(ans);
+  GstState old_state, new_state, pending;
+
+  gst_message_parse_state_changed(Message_val(_msg), &old_state, &new_state, &pending);
+
+  ans = caml_alloc_tuple(3);
+  Store_field(ans, 0, val_of_state(old_state));
+  Store_field(ans, 1, val_of_state(new_state));
+  Store_field(ans, 2, val_of_state(pending));
+
+  CAMLreturn(ans);
+}
+
 CAMLprim value ocaml_gstreamer_message_parse_tag(value _msg)
 {
   CAMLparam1(_msg);
@@ -406,10 +517,16 @@ CAMLprim value ocaml_gstreamer_message_parse_tag(value _msg)
       for (j = 0; j < n; j++)
         {
           val = gst_tag_list_get_value_index(tags, tag, j);
-          if (G_VALUE_HOLDS_STRING(val))
+          if (G_VALUE_HOLDS_STRING(val)) {
               s = caml_copy_string(g_value_get_string(val));
-          else
-            {
+            }
+          else if (GST_VALUE_HOLDS_DATE_TIME(val)) {
+              GstDateTime *dt = g_value_get_boxed(val);
+              gchar *dt_str = gst_date_time_to_iso8601_string(dt);
+              s = caml_copy_string(dt_str);
+              g_free(dt_str);
+            }
+          else {
               //TODO: better typed handling of non-string values?
               char *vc = g_strdup_value_contents(val);
               s = caml_copy_string(vc);
@@ -427,21 +544,74 @@ CAMLprim value ocaml_gstreamer_message_parse_tag(value _msg)
   CAMLreturn(ans);
 }
 
+/**** Main Loop ****/
+
+#define Loop_val(v) (*(GMainLoop**)Data_custom_val(v))
+
+static void finalize_loop(value v)
+{
+  GMainLoop *loop = Loop_val(v);
+  g_main_loop_unref(loop);
+}
+
+static struct custom_operations loop_ops =
+  {
+    "ocaml_gstreamer_loop",
+    finalize_loop,
+    custom_compare_default,
+    custom_hash_default,
+    custom_serialize_default,
+    custom_deserialize_default
+  };
+
+CAMLprim value ocaml_gstreamer_loop_create(value unit)
+{
+  CAMLparam0();
+  CAMLlocal1(ans);
+  GMainLoop* loop = g_main_loop_new(NULL, FALSE);
+
+  if (!loop)
+    caml_raise_out_of_memory();
+
+  ans = caml_alloc_custom(&loop_ops, sizeof(GMainLoop*), 0, 1);
+  Loop_val(ans) = loop;
+  
+  CAMLreturn(ans);
+}
+
+CAMLprim value ocaml_gstreamer_loop_run(value l)
+{
+  CAMLparam1(l);
+  GMainLoop* loop = Loop_val(l);
+
+  caml_release_runtime_system();
+  g_main_loop_run(loop);
+  caml_acquire_runtime_system();
+
+  CAMLreturn(Val_unit);
+}
+
+CAMLprim value ocaml_gstreamer_loop_quit(value l)
+{
+  CAMLparam1(l);
+  GMainLoop* loop = Loop_val(l);
+  
+  caml_release_runtime_system();
+  g_main_loop_quit(loop);
+  caml_acquire_runtime_system();
+
+  CAMLreturn(Val_unit);
+}
+
+
 /**** Bus ****/
 
-typedef struct {
-  GstBus *bus;
-  value element;
-} bus_t;
-
-#define Bus_data_val(v) (*(bus_t**)Data_custom_val(v))
-#define Bus_val(v) (Bus_data_val(v)->bus)
+#define Bus_val(v) (*(GstBus**)Data_custom_val(v))
 
 static void finalize_bus(value v)
 {
-  bus_t *bus = Bus_data_val(v);
-  caml_remove_global_root(&bus->element);
-  free(bus);
+  GstBus *bus = Bus_val(v);
+  gst_object_unref(bus);
 }
 
 static struct custom_operations bus_ops =
@@ -454,41 +624,29 @@ static struct custom_operations bus_ops =
     custom_deserialize_default
   };
 
-static value value_of_bus(GstBus *b)
-{
-  if (!b) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
-  value ans = caml_alloc_custom(&bus_ops, sizeof(bus_t*), 0, 1);
-  bus_t *bus = malloc(sizeof(bus));
-  bus->bus = b;
-  bus->element = 0;
-  caml_register_global_root(&bus->element);
-  Bus_data_val(ans) = bus;
-  return ans;
-}
-
 CAMLprim value ocaml_gstreamer_bus_of_element(value _e)
 {
   CAMLparam1(_e);
   CAMLlocal1(ans);
   GstElement *e = Element_val(_e);
-  ans = value_of_bus(GST_ELEMENT_BUS(e));
-  /* We keep a ref on this element so that the bus does not get invalidated by
-     the garbage collection of e */
-  Bus_data_val(ans)->element = _e;
+
+  ans = caml_alloc_custom(&bus_ops, sizeof(GstBus*), 0, 1);
+  Bus_val(ans) = gst_element_get_bus(e);
+
   CAMLreturn(ans);
 }
 
 CAMLprim value ocaml_gstreamer_bus_pop_filtered(value _bus, value _filter)
 {
   CAMLparam2(_bus, _filter);
-  CAMLlocal1(ans);
+  CAMLlocal2(ans,ret);
   GstBus *bus = Bus_val(_bus);
   GstMessageType filter = 0;
   GstMessage *msg;
   int i;
 
   for(i = 0; i < Wosize_val(_filter); i++)
-    filter |= message_type_of_int(Field(_filter, i));
+    filter |= message_type_of_int(Int_val(Field(_filter, i)));
 
   caml_release_runtime_system();
   msg = gst_bus_pop_filtered(bus, filter);
@@ -498,31 +656,40 @@ CAMLprim value ocaml_gstreamer_bus_pop_filtered(value _bus, value _filter)
     ans = Val_int(0);
   else
     {
+      value_of_message(msg,ret);
       ans = caml_alloc_tuple(1);
-      Store_field(ans, 0, value_of_message(msg));
+      Store_field(ans, 0, ret);
     }
 
   CAMLreturn(ans);
 }
 
-CAMLprim value ocaml_gstreamer_bus_timed_pop_filtered(value _bus, value _filter)
+CAMLprim value ocaml_gstreamer_bus_timed_pop_filtered(value _bus, value _timeout, value _filter)
 {
-  CAMLparam2(_bus, _filter);
+  CAMLparam3(_bus, _timeout, _filter);
   CAMLlocal1(ans);
   GstBus *bus = Bus_val(_bus);
+  GstClockTime timeout = GST_CLOCK_TIME_NONE;
   GstMessageType filter = 0;
   GstMessage *msg;
   int i;
 
+  if (Is_block(_timeout)) {
+    timeout = (GstClockTime)Int64_val(Field(_timeout, 0));
+  }
+
   for(i = 0; i < Wosize_val(_filter); i++)
-    filter |= message_type_of_int(Field(_filter, i));
+    filter |= message_type_of_int(Int_val(Field(_filter, i)));
 
   caml_release_runtime_system();
-  msg = gst_bus_timed_pop_filtered(bus, GST_CLOCK_TIME_NONE, filter);
+  msg = gst_bus_timed_pop_filtered(bus, timeout, filter);
   caml_acquire_runtime_system();
 
   if (!msg) caml_raise_constant(*caml_named_value("gstreamer_exn_timeout"));
-  CAMLreturn(value_of_message(msg));
+
+  value_of_message(msg,ans);
+
+  CAMLreturn(ans);
 }
 
 /***** Bin ******/
@@ -537,11 +704,13 @@ CAMLprim value ocaml_gstreamer_bin_add(value _bin, value _e)
   gboolean ret;
 
   caml_release_runtime_system();
+  /* gst_bin_add takes ownership but we want to keep it alive 
+   * and clean it with the Gc. */
   gst_object_ref(e);
   ret = gst_bin_add(bin, e);
   caml_acquire_runtime_system();
 
-  if (!ret) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
+  if (!ret) caml_raise_constant(*caml_named_value("gstreamer_exn_failed"));
   CAMLreturn(Val_unit);
 }
 
@@ -554,7 +723,9 @@ CAMLprim value ocaml_gstreamer_bin_get_by_name(value _bin, value _name)
 
   e = gst_bin_get_by_name(bin, String_val(_name));
 
-  CAMLreturn(value_of_element(e));
+  value_of_element(e,ans);
+
+  CAMLreturn(ans);
 }
 
 /***** Pipeline *****/
@@ -569,26 +740,27 @@ CAMLprim value ocaml_gstreamer_pipeline_create(value s)
 
   e = gst_pipeline_new(String_val(s));
 
-  ans = value_of_element(e);
+  value_of_element(e,ans);
+
   CAMLreturn(ans);
 }
 
 CAMLprim value ocaml_gstreamer_pipeline_parse_launch(value s)
 {
   CAMLparam1(s);
-  CAMLlocal1(ans);
+  CAMLlocal2(ans,_err);
   GError *err = NULL;
   GstElement *e;
 
   e = gst_parse_launch(String_val(s), &err);
-  if (err)
+  if (err && e == NULL)
     {
-      value s = caml_copy_string(err->message);
-      if (e) { gst_object_unref(e); }
+      _err = caml_copy_string(err->message);
       g_error_free(err);
-      caml_raise_with_arg(*caml_named_value("gstreamer_exn_failure_msg"), s);
+      caml_raise_with_arg(*caml_named_value("gstreamer_exn_error"), _err);
     }
-  ans = value_of_element(e);
+
+  value_of_element(e,ans);
 
   CAMLreturn(ans);
 }
@@ -613,64 +785,73 @@ static struct custom_operations buffer_ops =
     custom_deserialize_default
   };
 
-static value value_of_buffer(GstBuffer *b)
-{
-  if (!b) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
-  value ans = caml_alloc_custom(&buffer_ops, sizeof(GstElement*), 0, 1);
+#define value_of_buffer(b,ans) \
+  ans = caml_alloc_custom(&buffer_ops, sizeof(GstElement*), 0, 1); \
   Buffer_val(ans) = b;
-  return ans;
-}
+
+#define buffer_fill(gstbuf,data,ofs,len) \
+  GstMapInfo map; \
+  gboolean bret; \
+  caml_release_runtime_system(); \
+  bret = gst_buffer_map(gstbuf, &map, GST_MAP_WRITE); \
+  caml_acquire_runtime_system(); \
+  if(!bret) caml_raise_out_of_memory(); \
+  memcpy(map.data, (unsigned char*)data+ofs, len);\
+  caml_release_runtime_system(); \
+  gst_buffer_unmap(gstbuf, &map); \
+  caml_acquire_runtime_system();
+
 
 CAMLprim value ocaml_gstreamer_buffer_of_string(value s, value _off, value _len)
 {
   CAMLparam1(s);
+  CAMLlocal1(ans);
+
   int bufoff = Int_val(_off);
   int buflen = Int_val(_len);
   GstBuffer *gstbuf;
-  GstMapInfo map;
-  gboolean bret;
+  char *data;
 
   assert(buflen+bufoff <= caml_string_length(s));
 
   caml_release_runtime_system();
-  gstbuf = gst_buffer_new_and_alloc(buflen);
-  bret = gst_buffer_map(gstbuf, &map, GST_MAP_WRITE);
+  gstbuf = gst_buffer_new_allocate(NULL, buflen, NULL);
   caml_acquire_runtime_system();
 
-  if(!bret) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
-  memcpy(map.data, (unsigned char*)String_val(s)+bufoff, buflen);
+  if(!gstbuf) caml_raise_out_of_memory();
 
-  caml_release_runtime_system();
-  gst_buffer_unmap(gstbuf, &map);
-  caml_acquire_runtime_system();
+  data = String_val(s);
+  buffer_fill(gstbuf, data, bufoff, buflen);
 
-  CAMLreturn(value_of_buffer(gstbuf));
+  value_of_buffer(gstbuf,ans);
+
+  CAMLreturn(ans);
 }
 
 CAMLprim value ocaml_gstreamer_buffer_of_data(value _ba, value _off, value _len)
 {
   CAMLparam1(_ba);
+  CAMLlocal1(ans);
+
   int bufoff = Int_val(_off);
   int buflen = Int_val(_len);
   GstBuffer *gstbuf;
-  GstMapInfo map;
-  gboolean bret;
+  char *data;
 
   assert(buflen+bufoff <= Caml_ba_array_val(_ba)->dim[0]);
 
   caml_release_runtime_system();
-  gstbuf = gst_buffer_new_and_alloc(buflen);
-  bret = gst_buffer_map(gstbuf, &map, GST_MAP_WRITE);
+  gstbuf = gst_buffer_new_allocate(NULL, buflen, NULL);
   caml_acquire_runtime_system();
 
-  if(!bret) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
-  memcpy(map.data, (unsigned char*)Caml_ba_data_val(_ba)+bufoff, buflen);
+  if(!gstbuf) caml_raise_out_of_memory();
 
-  caml_release_runtime_system();
-  gst_buffer_unmap(gstbuf, &map);
-  caml_acquire_runtime_system();
+  data = Caml_ba_data_val(_ba);
+  buffer_fill(gstbuf, data, bufoff, buflen);
 
-  CAMLreturn(value_of_buffer(gstbuf));
+  value_of_buffer(gstbuf,ans);
+
+  CAMLreturn(ans);
 }
 
 CAMLprim value ocaml_gstreamer_buffer_set_presentation_time(value _buf, value _t)
@@ -710,6 +891,7 @@ CAMLprim value ocaml_gstreamer_buffer_set_duration(value _buf, value _t)
 
 typedef struct {
   GstAppSrc *appsrc;
+  value element;
   value need_data_cb; // Callback function
   gulong need_data_hid; // Callback handler ID
 } appsrc;
@@ -725,7 +907,7 @@ static void disconnect_need_data(appsrc *as)
     }
   if(as->need_data_cb)
     {
-      caml_remove_global_root(&as->need_data_cb);
+      caml_remove_generational_global_root(&as->need_data_cb);
       as->need_data_cb = 0;
     }
 }
@@ -734,6 +916,10 @@ static void finalize_appsrc(value v)
 {
   appsrc *as = Appsrc_val(v);
   disconnect_need_data(as);
+  if (as->element) {
+    caml_remove_generational_global_root(&as->element);
+    as->element = 0;
+  }
   free(as);
 }
 
@@ -747,58 +933,77 @@ static struct custom_operations appsrc_ops =
     custom_deserialize_default
   };
 
-static value value_of_appsrc(GstAppSrc *e)
-{
-  value ans = caml_alloc_custom(&appsrc_ops, sizeof(appsrc*), 0, 1);
-  appsrc *as = malloc(sizeof(appsrc));
-  as->appsrc = e;
-  as->need_data_cb = 0;
-  as->need_data_hid = 0;
-  Appsrc_val(ans) = as;
-  return ans;
-}
-
 CAMLprim value ocaml_gstreamer_appsrc_of_element(value _e)
 {
   CAMLparam1(_e);
+  CAMLlocal1(ans);
+
   GstElement *e = Element_val(_e);
-  CAMLreturn(value_of_appsrc(GST_APP_SRC(e)));
+
+  appsrc *as = malloc(sizeof(appsrc));
+  if (as == NULL)
+    caml_raise_out_of_memory();
+
+  as->appsrc = GST_APP_SRC(e);
+  as->need_data_cb = 0;
+  as->need_data_hid = 0;
+  as->element = _e;
+  caml_register_global_root(&as->element);
+
+  ans = caml_alloc_custom(&appsrc_ops, sizeof(appsrc*), 0, 1);
+  Appsrc_val(ans) = as;
+
+  CAMLreturn(ans);
 }
 
 CAMLprim value ocaml_gstreamer_appsrc_to_element(value _as)
 {
   CAMLparam1(_as);
+  CAMLlocal1(ans);
   appsrc *as = Appsrc_val(_as);
-  CAMLreturn(value_of_element(GST_ELEMENT(as->appsrc)));
+  GstElement *e = GST_ELEMENT(as->appsrc);
+  value_of_element(e,ans);
+  CAMLreturn(ans);
 }
 
-CAMLprim value ocaml_gstreamer_appsrc_push_buffer_string(value _as, value _buf)
+CAMLprim value ocaml_gstreamer_appsrc_push_buffer_bytes_n(value _as, value _pres_time, value _dur, value _buf, value _ofs, value _len)
 {
-  CAMLparam2(_as, _buf);
-  int buflen = caml_string_length(_buf);
+  CAMLparam4(_as, _pres_time, _dur, _buf);
   appsrc *as = Appsrc_val(_as);
   GstBuffer *gstbuf;
-  GstMapInfo map;
   GstFlowReturn ret;
-  gboolean bret;
+  int64_t pres_time = Int64_val(_pres_time);
+  int64_t dur = Int64_val(_dur);
+  unsigned char *data;
 
   caml_release_runtime_system();
-  gstbuf = gst_buffer_new_and_alloc(buflen);
-  bret = gst_buffer_map(gstbuf, &map, GST_MAP_WRITE);
+  gstbuf = gst_buffer_new_allocate(NULL, Int_val(_len), NULL);
   caml_acquire_runtime_system();
 
-  if(!bret) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
-  memcpy(map.data, (unsigned char*)String_val(_buf), buflen);
+  if(!gstbuf) caml_raise_out_of_memory();
+
+  if (pres_time >= 0)
+    gstbuf->pts = pres_time;
+
+  if (dur >= 0)
+    gstbuf->duration = dur;
+
+  data = Bytes_val(_buf);
+  buffer_fill(gstbuf, data, Int_val(_ofs), Int_val(_len));
 
   caml_release_runtime_system();
-  gst_buffer_unmap(gstbuf, &map);
-  ret = gst_app_src_push_buffer(as->appsrc, gstbuf);
+  ret = gst_app_src_push_buffer(as->appsrc,gstbuf);
   caml_acquire_runtime_system();
 
-  if (ret != GST_FLOW_OK) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
+  if (ret != GST_FLOW_OK) caml_raise_constant(*caml_named_value("gstreamer_exn_failed"));
   CAMLreturn(Val_unit);
 }
 
+CAMLprim value ocaml_gstreamer_appsrc_push_buffer_bytes_b(value * argv, int argn)
+{
+  return ocaml_gstreamer_appsrc_push_buffer_bytes_n(argv[0], argv[1], argv[2], argv[3], argv[4], argv[5]);
+}
+
 CAMLprim value ocaml_gstreamer_appsrc_push_buffer(value _as, value _buf)
 {
   CAMLparam2(_as, _buf);
@@ -807,24 +1012,59 @@ CAMLprim value ocaml_gstreamer_appsrc_push_buffer(value _as, value _buf)
   GstFlowReturn ret;
 
   caml_release_runtime_system();
-  /* The reference will be eaten by push_buffer */
-  gst_buffer_ref(gstbuf);
-  ret = gst_app_src_push_buffer(as->appsrc, gstbuf);
+  g_signal_emit_by_name(GST_ELEMENT(as->appsrc), "push-buffer", gstbuf, &ret);
   caml_acquire_runtime_system();
 
-  if (ret != GST_FLOW_OK) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
+  if (ret != GST_FLOW_OK) caml_raise_constant(*caml_named_value("gstreamer_exn_failed"));
   CAMLreturn(Val_unit);
 }
 
+CAMLprim value ocaml_gstreamer_appsrc_push_buffer_data_n(value _as, value _pres_time, value _dur, value _buf, value _ofs, value _len)
+{
+  CAMLparam4(_as, _pres_time, _dur, _buf);
+  appsrc *as = Appsrc_val(_as);
+  GstBuffer *gstbuf;
+  GstFlowReturn ret;
+  int64_t pres_time = Int64_val(_pres_time);
+  int64_t dur = Int64_val(_dur);
+  char *data;
+
+  caml_release_runtime_system();
+  gstbuf = gst_buffer_new_allocate(NULL, Int_val(_len), NULL);
+  caml_acquire_runtime_system();
+
+  if(!gstbuf) caml_raise_out_of_memory();
+
+  if (pres_time >= 0)
+    gstbuf->pts = pres_time;
+
+  if (dur >= 0)
+    gstbuf->duration = dur;
+
+  data = Caml_ba_data_val(_buf);
+  buffer_fill(gstbuf, data, Int_val(_ofs), Int_val(_len));
+
+  caml_release_runtime_system();
+  ret = gst_app_src_push_buffer(as->appsrc,gstbuf);
+  caml_acquire_runtime_system();
+
+  if (ret != GST_FLOW_OK) caml_raise_constant(*caml_named_value("gstreamer_exn_failed"));
+  CAMLreturn(Val_unit);
+}
+
+CAMLprim value ocaml_gstreamer_appsrc_push_buffer_data_b(value * argv, int argn)
+{
+  return ocaml_gstreamer_appsrc_push_buffer_data_n(argv[0], argv[1], argv[2], argv[3], argv[4], argv[5]);
+}
+
 static void appsrc_need_data_cb(GstAppSrc *gas, guint length, gpointer user_data)
 {
   appsrc *as = (appsrc*)user_data;
 
-  caml_c_thread_register();
+  ocaml_gstreamer_register_thread();
   caml_acquire_runtime_system();
   caml_callback(as->need_data_cb, Val_int(length));
   caml_release_runtime_system();
-  caml_c_thread_unregister();
 }
 
 CAMLprim value ocaml_gstreamer_appsrc_connect_need_data(value _as, value f)
@@ -833,14 +1073,14 @@ CAMLprim value ocaml_gstreamer_appsrc_connect_need_data(value _as, value f)
   appsrc *as = Appsrc_val(_as);
   disconnect_need_data(as);
 
-  caml_register_global_root(&as->need_data_cb);
+  as->need_data_cb = f;
+  caml_register_generational_global_root(&as->need_data_cb);
 
   caml_release_runtime_system();
-  as->need_data_cb = f;
   as->need_data_hid = g_signal_connect(as->appsrc, "need-data", G_CALLBACK(appsrc_need_data_cb), as);
   caml_acquire_runtime_system();
 
-  if(!as->need_data_hid) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
+  if(!as->need_data_hid) caml_raise_constant(*caml_named_value("gstreamer_exn_failed"));
   CAMLreturn(Val_unit);
 }
 
@@ -854,7 +1094,7 @@ CAMLprim value ocaml_gstreamer_appsrc_end_of_stream(value _as)
   g_signal_emit_by_name(as->appsrc, "end-of-stream", &ret);
   caml_acquire_runtime_system();
 
-  if(ret != GST_FLOW_OK) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
+  if(ret != GST_FLOW_OK) caml_raise_constant(*caml_named_value("gstreamer_exn_failed"));
   CAMLreturn(Val_unit);
 }
 
@@ -875,6 +1115,7 @@ CAMLprim value ocaml_gstreamer_appsrc_set_format(value _as, value _fmt)
 
 typedef struct {
   GstAppSink *appsink;
+  value element;
   value new_sample_cb; // Callback function
   gulong new_sample_hid; // Callback handler ID
 } appsink;
@@ -890,7 +1131,7 @@ static void disconnect_new_sample(appsink *as)
     }
   if(as->new_sample_cb)
     {
-      caml_remove_global_root(&as->new_sample_cb);
+      caml_remove_generational_global_root(&as->new_sample_cb);
       as->new_sample_cb = 0;
     }
 }
@@ -899,6 +1140,10 @@ static void finalize_appsink(value v)
 {
   appsink *as = Appsink_val(v);
   disconnect_new_sample(as);
+  if (as->element) {
+    caml_remove_generational_global_root(&as->element);
+    as->element = 0;
+  }
   free(as);
 }
 
@@ -912,22 +1157,28 @@ static struct custom_operations appsink_ops =
     custom_deserialize_default
   };
 
-static value value_of_appsink(GstAppSink *e)
-{
-  value ans = caml_alloc_custom(&appsink_ops, sizeof(appsink*), 0, 1);
-  appsink *as = malloc(sizeof(appsink));
-  as->appsink = e;
-  as->new_sample_cb = 0;
-  as->new_sample_hid = 0;
-  Appsink_val(ans) = as;
-  return ans;
-}
-
 CAMLprim value ocaml_gstreamer_appsink_of_element(value _e)
 {
   CAMLparam1(_e);
+  CAMLlocal1(ans);
+
   GstElement *e = Element_val(_e);
-  CAMLreturn(value_of_appsink(GST_APP_SINK(e)));
+  appsink *as = malloc(sizeof(appsink));
+
+  if (as == NULL)
+    caml_raise_out_of_memory();
+
+  as->appsink = GST_APP_SINK(e);
+  as->element = _e;
+  as->new_sample_cb = 0;
+  as->new_sample_hid = 0;
+  as->element = _e;
+  caml_register_generational_global_root(&as->element);
+
+  ans = caml_alloc_custom(&appsink_ops, sizeof(appsink*), 0, 1);
+  Appsink_val(ans) = as;
+
+  CAMLreturn(ans);
 }
 
 CAMLprim value ocaml_gstreamer_appsink_emit_signals(value _as)
@@ -942,7 +1193,7 @@ CAMLprim value ocaml_gstreamer_appsink_emit_signals(value _as)
   CAMLreturn(Val_unit);
 }
 
-CAMLprim value ocaml_gstreamer_appsink_pull_buffer(value _as)
+CAMLprim value ocaml_gstreamer_appsink_pull_buffer(value _as, value string_mode)
 {
   CAMLparam1(_as);
   CAMLlocal1(ans);
@@ -962,24 +1213,29 @@ CAMLprim value ocaml_gstreamer_appsink_pull_buffer(value _as)
       if (gst_app_sink_is_eos(as->appsink))
         caml_raise_constant(*caml_named_value("gstreamer_exn_eos"));
       else
-        caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
+        caml_raise_constant(*caml_named_value("gstreamer_exn_stopped"));
     }
 
   caml_release_runtime_system();
   gstbuf = gst_sample_get_buffer(gstsample);
   caml_acquire_runtime_system();
 
-  if (!gstbuf) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
+  if (!gstbuf) caml_raise_out_of_memory();
 
   caml_release_runtime_system();
   ret = gst_buffer_map(gstbuf, &map, GST_MAP_READ);
   caml_acquire_runtime_system();
 
-  if (!ret) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
+  if (!ret) caml_raise_out_of_memory();
 
   len = map.size;
-  ans = caml_ba_alloc(CAML_BA_C_LAYOUT | CAML_BA_UINT8, 1, NULL, &len);
-  memcpy(Caml_ba_data_val(ans), map.data, len);
+  if (string_mode == Val_false) {
+    ans = caml_ba_alloc(CAML_BA_C_LAYOUT | CAML_BA_UINT8, 1, NULL, &len);
+    memcpy(Caml_ba_data_val(ans), map.data, len);
+  } else {
+    ans = caml_alloc_string(len);
+    memcpy(String_val(ans), map.data, len);
+  }
 
   caml_release_runtime_system();
   gst_buffer_unmap(gstbuf, &map);
@@ -989,21 +1245,6 @@ CAMLprim value ocaml_gstreamer_appsink_pull_buffer(value _as)
   CAMLreturn(ans);
 }
 
-/* TODO: optimized version? */
-CAMLprim value ocaml_gstreamer_appsink_pull_buffer_string(value _as)
-{
-  CAMLparam1(_as);
-  CAMLlocal2(ba,ans);
-  int len;
-
-  ba = ocaml_gstreamer_appsink_pull_buffer(_as);
-  len = Caml_ba_array_val(ba)->dim[0];
-  ans = caml_alloc_string(len);
-  memcpy(String_val(ans), Caml_ba_data_val(ba), len);
-
-  CAMLreturn(ans);
-}
-
 CAMLprim value ocaml_gstreamer_appsink_is_eos(value _as)
 {
   CAMLparam1(_as);
@@ -1021,11 +1262,10 @@ static GstFlowReturn appsink_new_sample_cb(GstAppSink *gas, gpointer user_data)
 {
   appsink *as = (appsink*)user_data;
 
-  caml_c_thread_register();
+  ocaml_gstreamer_register_thread();
   caml_acquire_runtime_system();
   caml_callback(as->new_sample_cb, Val_unit);
   caml_release_runtime_system();
-  caml_c_thread_unregister();
 
   return GST_FLOW_OK;
 }
@@ -1036,14 +1276,14 @@ CAMLprim value ocaml_gstreamer_appsink_connect_new_sample(value _as, value f)
   appsink *as = Appsink_val(_as);
   disconnect_new_sample(as);
 
-  caml_register_global_root(&as->new_sample_cb);
+  as->new_sample_cb = f;
+  caml_register_generational_global_root(&as->new_sample_cb);
 
   caml_release_runtime_system();
-  as->new_sample_cb = f;
   as->new_sample_hid = g_signal_connect(as->appsink, "new-sample", G_CALLBACK(appsink_new_sample_cb), as);
   caml_acquire_runtime_system();
 
-  if (!as->new_sample_hid) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
+  if (!as->new_sample_hid) caml_raise_constant(*caml_named_value("gstreamer_exn_failed"));
   CAMLreturn(Val_unit);
 }
 
@@ -1082,7 +1322,6 @@ static struct custom_operations caps_ops =
 
 static value value_of_caps(GstCaps *c)
 {
-  if (!c) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
   value ans = caml_alloc_custom(&caps_ops, sizeof(GstCaps*), 0, 1);
   Caps_val(ans) = c;
   return ans;
@@ -1172,12 +1411,9 @@ static void typefind_element_have_type_cb(GstElement *_typefind, guint probabili
   assert(_typefind);
   assert(caps);
 
-  //For some reason, we segfault if we register the C thread (I guess the implementation is monothreaded?)
-  /* caml_c_thread_register(); */
   caml_acquire_runtime_system();
   caml_callback2(tf->have_type_cb, Val_int(probability), value_of_caps(caps));
   caml_release_runtime_system();
-  /* caml_c_thread_unregister(); */
 }
 
 CAMLprim value ocaml_gstreamer_typefind_element_connect_have_type(value _tf, value f)
@@ -1193,7 +1429,7 @@ CAMLprim value ocaml_gstreamer_typefind_element_connect_have_type(value _tf, val
   tf->have_type_hid = g_signal_connect(tf->tf, "have-type", G_CALLBACK(typefind_element_have_type_cb), tf);
   caml_acquire_runtime_system();
 
-  if (!tf->have_type_hid) caml_raise_constant(*caml_named_value("gstreamer_exn_failure"));
+  if (!tf->have_type_hid) caml_raise_constant(*caml_named_value("gstreamer_exn_failed"));
   CAMLreturn(Val_unit);
 }
 
